# $Id: SQL.pm 51114 2023-11-14 10:07:28Z wsl $
# $URL: https://svn.uvt.nl/its-id/trunk/sources/pwdmodifier-pg/libpwdmodifier/lib/UvT/PwdModifier/SQL.pm $
#
# Er wordt door deze app geschreven naar adb123, die wordt eens in het uur ofzo gelezen
# in het GLW en de waarden worden dan overgenomen naar
# GLW-INGRES tabel: adb113 en de Oracle tabel adb113r. (r is replica).
# (Is inhoudelelijk hetzelfde als adb313 in de mibase).
# Initiëel worden de passwords ook door DEA in adb113r geplaatst.
#
# IDM
# Vanuit FIM wordt usernamentabel adb413usernaam bijgewerkt en die gaat naar Mibase en Ingres.
# Dit is adb113(r) zonder passwords.
# describe adb413usernaam;

# Schrijven en lezen moet voortaan alleen in adb123

# describe adb113r@lnkdea.kub.nl;
# USERNAAM       NOT NULL VARCHAR2(8)
# ANR            NOT NULL CHAR(6)
# CAT            CHAR(1)
# PWSERV         VARCHAR2(30) initëel cleartext password
# PW_SHA         VARCHAR2(150)
# PW_GPG         VARCHAR2(1500)
# PW_RCRYPT      VARCHAR2(200)\
# USNUITG        datum uitgifte datepasswchanged			 VARCHAR2(8)
# USNVERV        datum vervallen					 VARCHAR2(8)
# USNSTAT        status, (moet 'a' zijn om naar ldap gestuurd te worden)
#
# describe adb123@lnkdea.kub.nl;
# Name			   Null?    Type
# ANR			   NOT NULL CHAR(6)
# USERNAME		   NOT NULL CHAR(8)
# PW_SHA_BEFORE		   NOT NULL VARCHAR2(150)
# PW_RCRYPT_BEFORE	   NOT NULL VARCHAR2(200)
# PW_SHA_AFTER		   NOT NULL VARCHAR2(150)
# PW_RCRYPT_AFTER	   NOT NULL VARCHAR2(200)
# CHANGE_TIME		   NOT NULL DATE
# CHANGED_BY		   VARCHAR2(40)
# TO_INGRES		   DATE

# describe adb131@lnkdea;
# ANR			   NOT NULL CHAR(6)
# EXTMAIL		   VARCHAR2(80)
# TOGLW			   CHAR(1)
# TOGLW_DATE	   DATE
# EPOCH_TIME	   NUMBER(10)
#
#
# describe adb132@lnkdea;
# ANR			   NOT NULL CHAR(6)
# SEQ			   NOT NULL NUMBER
# VRAAG			   NOT NULL VARCHAR2(256)
# ANTWOORD		   NOT NULL VARCHAR2(256)
#
#
# adb124@lnkdea.kub.nl (describe leidt tot de foutmelding: ERROR object KUB does not exist.
# ANR
# STATUS
# EPOCH_TIME
#
use strict;
use warnings FATAL => 'all';

package UvT::PwdModifier::SQL;

use Carp qw(cluck);
use DBI;
use Data::Dumper;
use List::Util qw(min);

use Spiffy -Base;

field debug => 0;
field 'db';
field 'dsn', -init => 'die "no database_dsn configured\n"';
field 'user', -init => 'die "no database_user configured\n"';
field 'pass' => '';
field 'result';

field 'pwd_read_table', -init => 'die "no database_pwd_read_table configured\n"';
field 'pwd_write_table', -init => 'die "no database_pwd_write_table configured\n"';
field 'mail_table', -init => 'die "no database_mail_table configured\n"';
field 'qa_table', -init => 'die "no database_qa_table configured\n"';

our $dbi_options = {
	AutoCommit => 0,
    PrintError => 0,
    RaiseError => 1,
    ShowErrorStatement => 1,
};

sub new {
	my $config = shift;
	my %options;
	while (my ($key, $value) = each %$config) {
		$options{$key} = $value if $key =~ s/^database_//;
	}
	$self = super(%options);

	my $db = DBI->connect($self->dsn, $self->user, $self->pass, $dbi_options);

	if ($db) {
		warn "DBIConnect: Succesfully connected with: $self->dsn";
	} else {
		warn "DBIConnect: Could not connect with: $self->dsn, $self->pass";
	}

	my $pwd_read_table = $self->pwd_read_table;
	my $pwd_write_table = $self->pwd_write_table;
	my $mail_table = $self->mail_table;
	my $qa_table = $self->qa_table;

	$self->{previousPasswords} = $db->prepare("SELECT pw_rcrypt_after FROM $pwd_read_table WHERE anr = ? ORDER BY change_time");
	$self->{previousSHAPasswords} = $db->prepare("SELECT pw_sha_after FROM $pwd_read_table WHERE anr = ? ORDER BY change_time");
	$self->{deletePasswordHistory} = $db->prepare("DELETE FROM $pwd_write_table WHERE anr = ?");
	$self->{initialPassword} = $db->prepare("SELECT pwserv FROM $pwd_read_table WHERE anr = ?");

	$self->{setPwd} = $db->prepare(
		"INSERT INTO $pwd_write_table (anr, username, pw_sha_before, pw_rcrypt_before, pw_sha_after, pw_rcrypt_after,
		changed_by, epoch_time, change_time) VALUES (?, ?, ?, ?, ?, ?, ?, ?, CURRENT_TIMESTAMP)"
	);

	$self->{getQA} = $db->prepare("SELECT vraag, antwoord FROM $qa_table WHERE anr = ? ORDER BY seq");
	$self->{setQA} = $db->prepare("INSERT INTO $qa_table (anr, seq, vraag, antwoord) VALUES (?, ?, ?, ?)");
	$self->{deleteQA} = $db->prepare("DELETE FROM $qa_table WHERE anr = ?");

	$self->{getMail} = $db->prepare("SELECT extmail, epoch_time FROM $mail_table WHERE anr = ?");
	$self->{setMail} = $db->prepare("INSERT INTO $mail_table (anr, extmail, epoch_time) VALUES (?, ?, ?)");
	$self->{deleteMail} = $db->prepare("DELETE FROM $mail_table WHERE anr = ?");

	$self->{cleansePasswordHistory} = $db->prepare("DELETE FROM $pwd_write_table WHERE anr NOT IN (SELECT anr FROM current_anrs)");
	$self->{cleanseQA} = $db->prepare("DELETE FROM $qa_table WHERE anr NOT IN (SELECT anr FROM current_anrs)");
	$self->{cleanseMail} = $db->prepare("DELETE FROM $mail_table WHERE anr NOT IN (SELECT anr FROM current_anrs)");

	$self->db($db);

	return $self;
}

sub commit {
	$self->db->commit;
}

sub rollback {
	$self->db->rollback;
}

sub query {
        my $name = shift;
        my $q = $self->{$name};
		my $debug = $self->debug;

		# pas op, soms is @_ leeg, daarom niet in de warn opnemen
		if ($name) {
			warn "running query: $name:" if $debug;
		} else {
			cluck "empty query!";
		}

		if ($_[0]) {
			my $str = '';
			map { $str .= ", $_" if $_} @_;
#			warn "parameters: $str";
		}
        my $result = $q->execute(@_);
		$self->result($result);
        return $q;
}

sub query_fetch {
	my $q = $self->query(@_);
	my $res; (@$res) = $q->fetch();
	$q->finish;
	return $res->[0]->[0];
}

sub query_fetchall {
	my $q = $self->query(@_);
	my $res = $q->fetchall_arrayref;
	$q->finish;
	return $res;
}

sub query_fetchall_hashref {
	my $field = shift;
	my $q = $self->query(@_);
	my $res = $q->fetchall_hashref($field);
	$q->finish;
	return $res;
}

sub query_fetchrow_hashref {
	my $q = $self->query(@_);
	my $res = $q->fetchrow_hashref();
	$q->finish;
	return $res;
}

sub query_do {
	my $q = $self->query(@_);
	my $affectedRows = $q->rows();
	$q->finish;
	return $affectedRows;
}

sub getQA {
	my $anr = shift;
	my $res = $self->query_fetchall('getQA', $anr);
	my $n = @$res;
	die "userError: Invalid number of questions ($n)"
		unless $n == 0 || $n == 2;
	return $res;
}

sub getQAStatus {
	my $anr = shift;
	my $res = $self->getQA($anr);
 	my $n = @$res;
	return $n?1:0;
}

sub removeDeletedUsers {
	my $currentAnrs = shift;
	my $numCurrentAnrs = @$currentAnrs;
	die "number of current ANRs ($numCurrentAnrs) is suspiciously low, bailing"
		if $numCurrentAnrs < 10000;

	my $db = $self->db;

	$db->prepare('CREATE TEMP TABLE current_anrs (anr TEXT) ON COMMIT DROP')->execute;

	my $prepareBatchInsert = sub {
		my $count = shift;
		my $values = join(', ', ('(?)') x $count);
		return $db->prepare("INSERT INTO current_anrs (anr) VALUES $values")
	};

	my $batchSize = 100;
	my $fullBatchQuery = $prepareBatchInsert->($batchSize);

	for (my $start = 0; $start < $numCurrentAnrs; $start += $batchSize) {
		my @batch = @$currentAnrs[$start..min($start + $batchSize, $numCurrentAnrs) - 1];
		my $query = @batch == $batchSize
			? $fullBatchQuery
			: $prepareBatchInsert->(scalar @batch);
		$query->execute(@batch);
	}

	$db->prepare('CREATE UNIQUE INDEX current_anrs_anr ON current_anrs (anr)')->execute;

	my $res = $self->query_do('cleansePasswordHistory');
	warn "cleansePasswordHistory removed: $res accounts";

	$res = $self->query_do('cleanseQA');
	warn "cleanseQA removed: $res accounts";

	$res = $self->query_do('cleanseMail');
	warn "cleanseMail removed: $res accounts";
}

sub deletePasswordHistory {
	my $anr = shift;
	my $affectedRows = $self->query_do('deletePasswordHistory', $anr);
	warn "Deleted $affectedRows records for anr: $anr";
	$affectedRows;
}

sub setQA {
	my ($anr, $q1, $a1, $q2, $a2) = @_;
	$self->query_do('deleteQA', $anr);
	$self->query_do('setQA', $anr, 1, $q1, $a1) if ($q1 && $a1);
	$self->query_do('setQA', $anr, 2, $q2, $a2) if ($q2 && $a2);
}

sub getMail {
	my $anr = shift;
	my $res = $self->query_fetchall('getMail', $anr);
	my $n = @$res;
	die "userError: Invalid number of mail addresses ($n)"
		if $n > 1;
	my $mail = $res->[0]->[0];
	return $mail || 0;
}

sub getQAMModified {
	my $anr = shift;
	my $res = $self->query_fetchall('getMail', $anr);
	my $list = @$res[0];
	my $time = 0;
	if ($list) {
		$time = $list->[1] if $list->[1];
	}
	return $time;
}

sub setMail {
	my ($anr, $mail) = @_;
	$self->query_do('deleteMail', $anr);
	$self->query_do('setMail', $anr, $mail, time);
}

sub setPwd {
	$self->query_do('setPwd', @_);
}

sub previousSHAPasswords {
	my $anr = shift;
	my $res = $self->query_fetchall('previousSHAPasswords', $anr);
	return [map { my $r = $_->[0]; $r } @$res];
}

sub getCurrentSHAPassword {
	my $anr = shift;
	my $a = $self->previousSHAPasswords($anr);
	return $a->[-1];
}

sub previousPasswords {
	my $anr = shift;
	my $res = $self->query_fetchall('previousPasswords', $anr);
 	return [map { my $r = $_->[0]; $r =~ s/^\{rcrypt\}//; $r } @$res];
}

sub getInitialPassword {
	my $anr = shift;
	my $res = $self->query_fetchall('initialPassword', $anr);
	my $n = @$res;
	return $res->[0][0];
}

#sub getLatestRcrtypPassword {
#	my $anr = shift;
#	my $res = $self->query_fetchall('getLatestRcrtypPassword', $anr);
# 	return [map { my $r = $_->[0]; $r =~ s/^\{rcrypt\}//; $r } @$res];
#
#}

sub getCurrentRcryptPassword {
	my $anr = shift;
	my $a = $self->previousPasswords($anr);
	return $a->[-1];
}
